#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

static double tic(void) { 
    return (clock()/ (double) CLOCKS_PER_SEC); 
}

static double toc(double t) { 
    double s = tic(); 
    if(s<t) return(0.0);
    return(s-t); 
}


double **createMatrix(int nr, int nc) {
    int i;
    double **mat;
    
    // Reservamos memoria
    mat = (double **) malloc( (nr)*sizeof(double *));
    if(mat==NULL) return(NULL);
    mat[0] = (double *) malloc( nr*nc*sizeof(double));
    
    if(mat[0]==NULL) return(NULL);
    
    for(i=1; i<nr; ++i)
        mat[i] = mat[i-1] + nc;
    
    return(mat);
}

// Libera la memoria del arreglo bidimensional
void freeMatrix(double **mat) {
    free(mat[0]);
    free(mat);
}


double **createMatrix2(int nr, int nc) {
    int i;
    double **mat;
    
    // Reservamos memoria
    mat = (double **) malloc( nr*sizeof(double *));
    if(mat==NULL) return(NULL);
    
    for(i=0; i<nr; ++i)
        mat[i] = (double *) malloc( nc*sizeof(double));
    
    return(mat);
}

// Libera la memoria del arreglo bidimensional
void freeMatrix2(double **mat, int nr) {
    for(int i=0; i<nr; ++i)
        free(mat[i]);
    free(mat);
}


// Calcula el producto de las matrices  A y B, y el resultado se almacena
// en C
void producto1(int rowsA, int colsA, double **A, int colsB, double **B, 
               double **C) {
    int i, j, k;

    for(i=0; i<rowsA; ++i)
        for(j=0; j<colsB; ++j) {
            C[i][j] = 0;
            for(k=0; k<colsA; ++k)
                C[i][j] += A[i][k]*B[k][j];
        }
}


// Calcula el producto de las matrices  A y B, y el resultado se almacena
// en C
void producto2(int rowsA, int colsA, double **A, int colsB, double **B, 
               double **C) {
    int     i, j, k;
    double  dtmp;

    for(i=0; i<rowsA; ++i)
        for(j=0; j<colsB; ++j) {
            dtmp = 0;
            for(k=0; k<colsA; ++k)
                dtmp += A[i][k]*B[k][j];
            C[i][j] = dtmp;
        }
}


// Calcula el producto de las matrices  A y B, y el resultado se almacena
// en C
void producto3(int rowsA, int colsA, double **A, int colsB, double **B, 
               double **C) {
    int     i, j, k;
    double  dtmp, *ptrA, *ptrB, *ptrC;

    for(i=0; i<rowsA; ++i) {
        ptrC = C[i];
        for(j=0; j<colsB; ++j, ++ptrC) {
            dtmp = 0;
            ptrA = A[i];
            ptrB = B[0]+j;
            for(k=0; k<colsA; ++k, ++ptrA, ptrB+=colsB)
                dtmp += *ptrA * *ptrB;
            *ptrC = dtmp;
        }
    }
}

// Calcula el producto de las matrices  mres = m1*m2
void producto4(int rows1, int cols1, int cols2, double *m1, double *m2, double *mres) {
    int     i, j, k;
    double  dval, *paux, *pm1, *pm2=m2, *pm3=mres;

    // inicializacion
    paux = m1;
    for(j=0; j<cols2; ++j) {
        dval  = *pm2;
        pm1   = paux;
        for(i=0; i<rows1; ++i, ++pm1, ++pm3) *pm3 = (*pm1)*dval;
        pm2  += cols1;
    }
    // suma del resto de las contribuciones
    for(k=1, paux=m1+rows1; k<cols1; ++k, paux += rows1) {
        pm3 = mres;
        pm2 = m2 + k;
        for(j=0; j<cols2; ++j) {
            dval  = *pm2;
            pm1   = paux;
            for(i=0; i<rows1; ++i, ++pm1, ++pm3) *pm3 += (*pm1)*dval;
            pm2  += cols1;
        }
    }
}

// Esta funcion no calcula el producto de dos matrices
void producto5(int rowsA, int colsA, double **A, int colsB, double **B, 
               double **C) {
    int     i, j, k;
    double  dtmp, *ptrA, *ptrB, *ptrC;

    ptrC = C[0];
    for(i=0; i<rowsA; ++i) {
        ptrB = B[0];
        for(j=0; j<colsB; ++j, ++ptrC) {
            dtmp = 0;
            ptrA = A[i];
            for(k=0; k<colsA; ++k, ++ptrA, ++ptrB)
                dtmp += *ptrA * *ptrB;
            *ptrC = dtmp;
        }
    }
}



int main(int argc, char **argv) {
    double  **A1, **B1, **C1, **A2, **B2, **C2, tini, tfin;
    int       n=5;

    if(argc>0) n = atoi(argv[1]);

    printf("Creando 3 matrices de tama~no %d \n", n);
    printf("**  reservando un bloque de memoria **\n");
    A1 = createMatrix(n, n);
    B1 = createMatrix(n, n);
    C1 = createMatrix(n, n);

    tini = tic();
    producto1(n, n, A1, n, B1, C1);
    tfin = toc(tini);
    printf("Producto 1. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto2(n, n, A1, n, B1, C1);
    tfin = toc(tini);
    printf("Producto 2. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto3(n, n, A1, n, B1, C1);
    tfin = toc(tini);
    printf("Producto 3. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto4(n, n, n, A1[0], B1[0], C1[0]);
    tfin = toc(tini);
    printf("Producto 4. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto5(n, n, A1, n, B1, C1);
    tfin = toc(tini);
    printf("Producto 5. Tiempo final = %e  segundos\n", tfin);

    freeMatrix(A1);
    freeMatrix(B1);
    freeMatrix(C1);

    printf("-----------------------------------\n");
    printf("Creando 3 matrices de tama~no %d \n", n);
    printf("** reservando varios bloques de memoria **\n");
    A2 = createMatrix2(n, n);
    B2 = createMatrix2(n, n);
    C2 = createMatrix2(n, n);

    tini = tic();
    producto1(n, n, A2, n, B2, C2);
    tfin = toc(tini);
    printf("Producto 1. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto2(n, n, A2, n, B2, C2);
    tfin = toc(tini);
    printf("Producto 2. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto3(n, n, A2, n, B2, C2);
    tfin = toc(tini);
    printf("Producto 3. Tiempo final = %e  segundos\n", tfin);

    freeMatrix2(A2, n);
    freeMatrix2(B2, n);
    freeMatrix2(C2, n);


    return(0);
}


