#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

static double tic(void) { 
    return (clock()/ (double) CLOCKS_PER_SEC); 
}

static double toc(double t) { 
    double s = tic(); 
    if(s<t) return(0.0);
    return(s-t); 
}


double **createMatrix(int nr, int nc) {
    int i;
    double **mat;
    
    // Reservamos memoria
    mat = (double **) malloc( (nr)*sizeof(double *));
    if(mat==NULL) return(NULL);
    mat[0] = (double *) malloc( nr*nc*sizeof(double));
    
    if(mat[0]==NULL) return(NULL);
    
    for(i=1; i<nr; ++i)
        mat[i] = mat[i-1] + nc;
    
    return(mat);
}

// Libera la memoria del arreglo bidimensional
void freeMatrix(double **mat) {
    free(mat[0]);
    free(mat);
}



double **createMatrix2(int nr, int nc) {
    int i;
    double **mat;
    
    // Reservamos memoria
    mat = (double **) malloc( nr*sizeof(double *));
    if(mat==NULL) return(NULL);
    
    for(i=0; i<nr; ++i)
        mat[i] = (double *) malloc( nc*sizeof(double));
    
    return(mat);
}

// Libera la memoria del arreglo bidimensional
void freeMatrix2(double **mat, int nr) {
    for(int i=0; i<nr; ++i)
        free(mat[i]);
    free(mat);
}

// Calcula  y = A*x. Se invierten los ciclos i y j de la funcion
// producto1. Esto deberia ser lo mas lento, aunque sean las mismas 
// operaciones.
void producto0(int m, int n, double **A, double *x, double *y) {
    int    i, j;

    for(i=0; i<m; ++i) 
        y[i] = A[i][0]*x[0];

    for(j=1; j<n; ++j) 
        for(i=0; i<m; ++i) 
            y[i] += A[i][j]*x[j];
    
}

// Calcula  y = A*x. Esta la forma estandar
void producto1(int m, int n, double **A, double *x, double *y) {
    int    i, j;

    for(i=0; i<m; ++i) {
        y[i] = 0;
        for(j=0; j<n; ++j) 
            y[i] += A[i][j]*x[j];
    }
}


// Calcula  y = A*x. Es igual que la funcion producto1, solo se
// cambia la forma de accesar a la informacion.
void producto2(int m, int n, double **A, double *x, double *y) {
    int    i, j;
    double dsum=0, *ptrA, *ptrx, *ptry=y;

    for(i=0; i<m; ++i, ++ptry) {
        dsum = 0;
        ptrx = x;
        ptrA = A[i];
        for(j=0; j<n; ++j, ++ptrx, ++ptrA) 
            dsum += (*ptrA) * (*ptrx);
        *ptry = dsum;
    }
}

// Calcula  y = A*x. Esta solo se puede usar cuando se reserva un solo
// bloque de memoria para la matriz.
void producto3(int m, int n, double **A, double *x, double *y) {
    int    i, j;
    double dsum=0, *ptrA=A[0], *ptrx, *ptry=y;

    for(i=0; i<m; ++i, ++ptry) {
        dsum = 0;
        ptrx = x;
        for(j=0; j<n; ++j, ++ptrx, ++ptrA) 
            dsum += (*ptrA) * (*ptrx);
        *ptry = dsum;
    }
}



int main(int argc, char **argv) {
    double  **A1, *x1, *y1, **A2, *x2, *y2, tini, tfin;
    int       n=5;

    if(argc>0) n = atoi(argv[1]);

    printf("Creando 1 matriz de tama~no %dx%d y un vector de tama~no %d\n", n, n, n);
    printf("**  reservando un bloque de memoria **\n");
    A1 = createMatrix(n, n);
    x1 = (double *) malloc( n*sizeof(double) );
    y1 = (double *) malloc( n*sizeof(double) );

    tini = tic();
    producto0(n, n, A1, x1, y1);
    tfin = toc(tini);
    printf("Producto 0. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto1(n, n, A1, x1, y1);
    tfin = toc(tini);
    printf("Producto 1. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto2(n, n, A1, x1, y1);
    tfin = toc(tini);
    printf("Producto 2. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto3(n, n, A1, x1, y1);
    tfin = toc(tini);
    printf("Producto 3. Tiempo final = %e  segundos\n", tfin);


    freeMatrix(A1);
    free(x1);
    free(y1);




    
    printf("-----------------------------------\n");
    printf("Creando 1 matriz de tama~no %dx%d y un vector de tama~no %d\n", n, n, n);
    printf("**  reservando varios bloques de memoria **\n");
    A2 = createMatrix2(n, n);
    x2 = (double *) malloc( n*sizeof(double) );
    y2 = (double *) malloc( n*sizeof(double) );

    tini = tic();
    producto0(n, n, A2, x2, y2);
    tfin = toc(tini);
    printf("Producto 0. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto1(n, n, A2, x2, y2);
    tfin = toc(tini);
    printf("Producto 1. Tiempo final = %e  segundos\n", tfin);

    tini = tic();
    producto2(n, n, A2, x2, y2);
    tfin = toc(tini);
    printf("Producto 2. Tiempo final = %e  segundos\n", tfin);

    freeMatrix2(A2, n);
    free(x2);
    free(y2);


    return(0);
}


